import re
from typing import Literal, TypedDict

from aiohttp import ClientSession

from yutto.exceptions import NoAccessPermissionError, UnSupportedTypeError
from yutto.media.codec import VideoCodec
from yutto.typing import AudioUrlMeta, AvId, BvId, CId, EpisodeId, MediaId, MultiLangSubtitle, SeasonId, VideoUrlMeta
from yutto.utils.console.logger import Logger
from yutto.utils.fetcher import Fetcher


class BangumiListItem(TypedDict):
    id: int
    name: str
    cid: CId
    episode_id: EpisodeId
    avid: AvId
    is_section: bool  # 是否属于专区


async def get_season_id_by_media_id(session: ClientSession, media_id: MediaId) -> SeasonId:
    home_url = "https://www.bilibili.com/bangumi/media/md{media_id}".format(media_id=media_id)
    season_id = SeasonId("")
    regex_season_id = re.compile(r'"param":{"season_id":(\d+),"season_type":\d+}')
    if match_obj := regex_season_id.search(await Fetcher.fetch_text(session, home_url)):
        season_id = match_obj.group(1)
    return SeasonId(str(season_id))


async def get_season_id_by_episode_id(session: ClientSession, episode_id: EpisodeId) -> SeasonId:
    home_url = "https://www.bilibili.com/bangumi/play/ep{episode_id}".format(episode_id=episode_id)
    season_id = SeasonId
    regex_season_id = re.compile(r'"id":\d+,"ssId":(\d+)')
    if match_obj := regex_season_id.search(await Fetcher.fetch_text(session, home_url)):
        season_id = match_obj.group(1)
    return SeasonId(str(season_id))


async def get_bangumi_title(session: ClientSession, season_id: SeasonId) -> str:
    play_url = "https://www.bilibili.com/bangumi/play/ss{season_id}".format(season_id=season_id)
    regex_title = re.compile(r'<a href=".+" target="_blank" title="(.*?)" class="media-title">(?P<title>.*?)</a>')
    if match_obj := regex_title.search(await Fetcher.fetch_text(session, play_url)):
        title = match_obj.group("title")
    else:
        title = "呐，我也不知道是什么标题呢～"
    return title


async def get_bangumi_list(session: ClientSession, season_id: SeasonId) -> list[BangumiListItem]:
    list_api = "http://api.bilibili.com/pgc/view/web/season?season_id={season_id}"
    resp_json = await Fetcher.fetch_json(session, list_api.format(season_id=season_id))
    result = resp_json["result"]
    section_episodes = []
    for section in result.get("section", []):
        section_episodes += section["episodes"]
    return [
        {
            "id": i + 1,
            "name": " ".join(
                [
                    "第{}话".format(item["title"]) if re.match(r"^\d*\.?\d*$", item["title"]) else item["title"],
                    item["long_title"],
                ]
            ),
            "cid": CId(str(item["cid"])),
            "episode_id": EpisodeId(str(item["id"])),
            "avid": BvId(item["bvid"]),
            "is_section": i >= len(result["episodes"]),
        }
        for i, item in enumerate(result["episodes"] + section_episodes)
    ]


async def get_bangumi_playurl(
    session: ClientSession, avid: AvId, episode_id: EpisodeId, cid: CId
) -> tuple[list[VideoUrlMeta], list[AudioUrlMeta]]:
    play_api = "https://api.bilibili.com/pgc/player/web/playurl?avid={aid}&bvid={bvid}&ep_id={episode_id}&cid={cid}&qn=125&fnver=0&fnval=16&fourk=1"
    codecid_map: dict[Literal[7, 12], VideoCodec] = {7: "avc", 12: "hevc"}

    async with session.get(
        play_api.format(**avid.to_dict(), cid=cid, episode_id=episode_id), proxy=Fetcher.proxy
    ) as resp:
        if not resp.ok:
            raise NoAccessPermissionError("无法下载该视频（cid: {cid}）".format(cid=cid))
        resp_json = await resp.json()
        if resp_json.get("result") is None:
            raise NoAccessPermissionError("无法下载该视频（cid: {cid}），原因：{msg}".format(cid=cid, msg=resp_json.get("message")))
        if resp_json["result"].get("dash") is None:
            raise UnSupportedTypeError("该视频（cid: {cid}）尚不支持 DASH 格式".format(cid=cid))
        if resp_json["result"]["is_preview"] == 1:
            Logger.warning("视频（cid: {cid}）是预览视频".format(cid=cid))
        return (
            [
                {
                    "url": video["base_url"],
                    "mirrors": video["backup_url"] if video["backup_url"] is not None else [],
                    "codec": codecid_map[video["codecid"]],
                    "width": video["width"],
                    "height": video["height"],
                    "quality": video["id"],
                }
                for video in resp_json["result"]["dash"]["video"]
            ],
            [
                {
                    "url": audio["base_url"],
                    "mirrors": audio["backup_url"] if audio["backup_url"] is not None else [],
                    "codec": "mp4a",
                    "width": 0,
                    "height": 0,
                    "quality": audio["id"],
                }
                for audio in resp_json["result"]["dash"]["audio"]
            ],
        )


async def get_bangumi_subtitles(session: ClientSession, avid: AvId, cid: CId) -> list[MultiLangSubtitle]:
    subtitile_api = "https://api.bilibili.com/x/player/v2?cid={cid}&aid={aid}&bvid={bvid}"
    subtitile_url = subtitile_api.format(**avid.to_dict(), cid=cid)
    subtitles_info = (await Fetcher.fetch_json(session, subtitile_url))["data"]["subtitle"]
    return [
        # fmt: off
        {
            "lang": sub_info["lan_doc"],
            "lines": (await Fetcher.fetch_json(session, "https:" + sub_info["subtitle_url"]))["body"]
        }
        for sub_info in subtitles_info["subtitles"]
        # fmt: on
    ]



import asyncio
import random
from typing import Any, Callable, Coroutine, Literal, Optional, TypeVar, Union
from urllib.parse import quote, unquote

import aiohttp
from aiohttp import ClientSession

from yutto.utils.console.logger import Logger
from yutto.utils.file_buffer import AsyncFileBuffer


class MaxRetryError(Exception):
    pass


T = TypeVar("T")


class MaxRetry:
    def __init__(self, max_retry: int = 2):
        self.max_retry = max_retry

    def __call__(self, connect_once: Callable[..., Coroutine[Any, Any, T]]) -> Callable[..., Coroutine[Any, Any, T]]:
        async def connect_n_times(*args: Any, **kwargs: Any) -> T:
            retry = self.max_retry + 1
            while retry:
                try:
                    return await connect_once(*args, **kwargs)
                except asyncio.TimeoutError as e:
                    Logger.warning("抓取超时，正在重试")
                finally:
                    retry -= 1
            raise MaxRetryError()

        return connect_n_times


class Fetcher:
    proxy: Optional[str] = None
    trust_env: bool = True
    headers: dict[str, str] = {
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/90.0.4430.85 Safari/537.36",
        "Referer": "https://www.bilibili.com",
    }
    cookies = {}

    @classmethod
    def set_proxy(cls, proxy: Union[Literal["no", "auto"], str]):
        if proxy == "auto":
            Fetcher.proxy = None
            Fetcher.trust_env = True
        elif proxy == "no":
            Fetcher.proxy = None
            Fetcher.trust_env = False
        else:
            Fetcher.proxy = proxy
            Fetcher.trust_env = False

    @classmethod
    def set_sessdata(cls, sessdata: str):
        # 先解码后编码是防止获取到的 SESSDATA 是已经解码后的（包含「,」）
        # 而番剧无法使用解码后的 SESSDATA
        Fetcher.cookies = {"SESSDATA": quote(unquote(sessdata))}

    @classmethod
    @MaxRetry(2)
    async def fetch_text(cls, session: ClientSession, url: str, encoding: Optional[str] = None) -> str:
        async with session.get(url, proxy=Fetcher.proxy) as resp:
            return await resp.text(encoding=encoding)

    @classmethod
    @MaxRetry(2)
    async def fetch_bin(cls, session: ClientSession, url: str) -> bytes:
        async with session.get(url, proxy=Fetcher.proxy) as resp:
            return await resp.read()

    @classmethod
    @MaxRetry(2)
    async def fetch_json(cls, session: ClientSession, url: str) -> Any:
        async with session.get(url, proxy=Fetcher.proxy) as resp:
            return await resp.json()

    @classmethod
    @MaxRetry(2)
    async def get_redirected_url(cls, session: ClientSession, url: str) -> str:
        async with session.get(url, proxy=Fetcher.proxy) as resp:
            return str(resp.url)

    @classmethod
    @MaxRetry(2)
    async def get_size(cls, session: ClientSession, url: str) -> Optional[int]:
        headers = session.headers.copy()
        headers["Range"] = "bytes=0-1"
        async with session.get(url, headers=headers, proxy=Fetcher.proxy) as resp:
            if resp.status == 206:
                return int(resp.headers["Content-Range"].split("/")[-1])
            else:
                return None

    @classmethod
    async def download_file_with_offset(
        cls,
        session: ClientSession,
        url: str,
        mirrors: list[str],
        file_buffer: AsyncFileBuffer,
        offset: int,
        size: Optional[int],
        stream: bool = True,
    ) -> None:
        done = False
        headers = session.headers.copy()
        url_pool = [url] + mirrors
        block_offset = 0
        while not done:
            try:
                url = random.choice(url_pool)
                headers["Range"] = "bytes={}-{}".format(
                    offset + block_offset, offset + size - 1 if size is not None else ""
                )
                async with session.get(
                    url, headers=headers, timeout=aiohttp.ClientTimeout(connect=5, sock_read=10), proxy=Fetcher.proxy
                ) as resp:
                    if stream:
                        while True:
                            # 如果直接用 1KiB 的话，会产生大量的块，消耗大量的 CPU 资源，
                            # 反而使得协程的优势不明显
                            # 而使用 1MiB 以上或者不使用流式下载方式时，由于分块太大，
                            # 导致进度条显示的实时速度并不准，波动太大，用户体验不佳，
                            # 因此取两者折中
                            chunk = await resp.content.read(2 ** 15)
                            if not chunk:
                                break
                            await file_buffer.write(chunk, offset + block_offset)
                            block_offset += len(chunk)
                    else:
                        chunk = await resp.read()
                        await file_buffer.write(chunk, offset + block_offset)
                        block_offset += len(chunk)
                # TODO: 是否需要校验总大小
                done = True

            except asyncio.TimeoutError as e:
                Logger.warning("文件 {} 下载超时，尝试重新连接...".format(file_buffer.file_path))

            except aiohttp.client_exceptions.ClientPayloadError as e:
                await asyncio.sleep(0.5)
                Logger.warning("文件 {} 下载出错，尝试重新连接...".format(file_buffer.file_path))



import json
import re
from typing import Literal, TypedDict

from aiohttp import ClientSession

from yutto.api.info import get_video_info
from yutto.exceptions import NoAccessPermissionError, UnSupportedTypeError
from yutto.media.codec import VideoCodec
from yutto.typing import AudioUrlMeta, AvId, CId, MultiLangSubtitle, VideoUrlMeta
from yutto.utils.fetcher import Fetcher


class AcgVideoListItem(TypedDict):
    id: int
    name: str
    cid: CId


async def get_acg_video_title(session: ClientSession, avid: AvId) -> str:
    return (await get_video_info(session, avid))["title"]


async def get_acg_video_list(session: ClientSession, avid: AvId) -> list[AcgVideoListItem]:
    list_api = "https://api.bilibili.com/x/player/pagelist?aid={aid}&bvid={bvid}&jsonp=jsonp"
    res_json = await Fetcher.fetch_json(session, list_api.format(**avid.to_dict()))
    return [
        # fmt: off
        {
            "id": i + 1,
            "name": item["part"],
            "cid": CId(str(item["cid"]))
        }
        # fmt: on
        for i, item in enumerate(res_json["data"])
    ]


async def get_acg_video_playurl(
    session: ClientSession, avid: AvId, cid: CId
) -> tuple[list[VideoUrlMeta], list[AudioUrlMeta]]:
    play_api = "https://api.bilibili.com/x/player/playurl?avid={aid}&bvid={bvid}&cid={cid}&qn=125&type=&otype=json&fnver=0&fnval=80&fourk=1"
    codecid_map: dict[Literal[7, 12], VideoCodec] = {7: "avc", 12: "hevc"}

    async with session.get(play_api.format(**avid.to_dict(), cid=cid), proxy=Fetcher.proxy) as resp:
        if not resp.ok:
            raise NoAccessPermissionError("无法下载该视频（cid: {cid}）".format(cid=cid))
        resp_json = await resp.json()
        if resp_json.get("data") is None:
            raise NoAccessPermissionError("无法下载该视频（cid: {cid}），原因：{msg}".format(cid=cid, msg=resp_json.get("message")))
        if resp_json["data"].get("dash") is None:
            raise UnSupportedTypeError("该视频（cid: {cid}）尚不支持 DASH 格式".format(cid=cid))
        return (
            [
                {
                    "url": video["base_url"],
                    "mirrors": video["backup_url"] if video["backup_url"] is not None else [],
                    "codec": codecid_map[video["codecid"]],
                    "width": video["width"],
                    "height": video["height"],
                    "quality": video["id"],
                }
                for video in resp_json["data"]["dash"]["video"]
            ]
            if resp_json["data"]["dash"]["video"]
            else [],
            [
                {
                    "url": audio["base_url"],
                    "mirrors": audio["backup_url"] if audio["backup_url"] is not None else [],
                    "codec": "mp4a",
                    "width": 0,
                    "height": 0,
                    "quality": audio["id"],
                }
                for audio in resp_json["data"]["dash"]["audio"]
            ]
            if resp_json["data"]["dash"]["audio"]
            else [],
        )


async def get_acg_video_subtitles(session: ClientSession, avid: AvId, cid: CId) -> list[MultiLangSubtitle]:
    subtitile_api = "https://api.bilibili.com/x/player.so?aid={aid}&bvid={bvid}&id=cid:{cid}"
    subtitile_url = subtitile_api.format(**avid.to_dict(), cid=cid)
    res_text = await Fetcher.fetch_text(session, subtitile_url)
    if subtitle_json_text_match := re.search(r"<subtitle>(.+)</subtitle>", res_text):
        subtitle_json = json.loads(subtitle_json_text_match.group(1))
        return [
            # fmt: off
            {
                "lang": sub_info["lan_doc"],
                "lines": (await Fetcher.fetch_json(session, "https:" + sub_info["subtitle_url"]))["body"]
            }
            for sub_info in subtitle_json["subtitles"]
            # fmt: on
        ]
    else:
        return []



from typing import Literal

VideoCodec = Literal["hevc", "avc"]
AudioCodec = Literal["mp4a"]

video_codec_priority_default: list[VideoCodec] = ["hevc", "avc"]
audio_codec_priority_default: list[AudioCodec] = ["mp4a"]


def gen_vcodec_priority(video_codec: VideoCodec) -> list[VideoCodec]:
    """生成视频编码优先级序列"""

    return ["hevc", "avc"] if video_codec == "hevc" else ["avc", "hevc"]


def gen_acodec_priority(audio_codec: AudioCodec) -> list[AudioCodec]:
    """生成音频编码优先级序列"""

    return ["mp4a"]



import argparse
import sys

import aiohttp

from yutto.api.acg_video import AcgVideoListItem, get_acg_video_list, get_acg_video_title
from yutto.api.bangumi import (
    get_bangumi_list,
    get_bangumi_title,
    get_season_id_by_episode_id,
    get_season_id_by_media_id,
)
from yutto.api.space import get_uploader_name, get_uploader_space_all_videos_avids
from yutto.cli.get import fetch_acg_video_data, fetch_bangumi_data
from yutto.exceptions import ErrorCode, HttpStatusError, NoAccessPermissionError, UnSupportedTypeError
from yutto.processor.downloader import process_video_download
from yutto.processor.filter import parse_episodes
from yutto.processor.urlparser import (
    regexp_acg_video_av,
    regexp_acg_video_bv,
    regexp_bangumi_ep,
    regexp_bangumi_md,
    regexp_bangumi_ss,
    regexp_space_all,
)
from yutto.typing import AId, AvId, BvId, EpisodeData, EpisodeId, MediaId, MId, SeasonId
from yutto.utils.console.logger import Badge, Logger
from yutto.utils.fetcher import Fetcher
from yutto.utils.functiontools.sync import sync


@sync
async def run(args: argparse.Namespace):
    async with aiohttp.ClientSession(
        headers=Fetcher.headers,
        cookies=Fetcher.cookies,
        trust_env=Fetcher.trust_env,
        timeout=aiohttp.ClientTimeout(total=5),
    ) as session:
        url: str = args.url
        url = await Fetcher.get_redirected_url(session, url)
        download_list: list[EpisodeData] = []
        if (
            (match_obj := regexp_bangumi_ep.match(url))
            or (match_obj := regexp_bangumi_ss.match(url))
            or (match_obj := regexp_bangumi_md.match(url))
        ):
            # 匹配为番剧
            if "episode_id" in match_obj.groupdict().keys():
                episode_id = EpisodeId(match_obj.group("episode_id"))
                season_id = await get_season_id_by_episode_id(session, episode_id)
            elif "season_id" in match_obj.groupdict().keys():
                season_id = SeasonId(match_obj.group("season_id"))
            else:
                media_id = MediaId(match_obj.group("media_id"))
                season_id = await get_season_id_by_media_id(session, media_id)
            title = await get_bangumi_title(session, season_id)
            Logger.custom(title, Badge("番剧", fore="black", back="cyan"))
            bangumi_list = await get_bangumi_list(session, season_id)
            # 如果没有 with_section 则不需要专区内容
            bangumi_list = list(filter(lambda item: args.with_section or not item["is_section"], bangumi_list))
            # 选集过滤
            episodes = parse_episodes(args.episodes, len(bangumi_list))
            bangumi_list = list(filter(lambda item: item["id"] in episodes, bangumi_list))
            for i, bangumi_item in enumerate(bangumi_list):
                Logger.status.set("正在努力解析第 {}/{} 个视频".format(i + 1, len(bangumi_list)))
                try:
                    episode_data = await fetch_bangumi_data(
                        session, bangumi_item["episode_id"], bangumi_item, args, {"title": title}, "{title}/{name}"
                    )
                except (NoAccessPermissionError, HttpStatusError, UnSupportedTypeError) as e:
                    Logger.error(e.message)
                    continue
                download_list.append(episode_data)

        elif (match_obj := regexp_acg_video_av.match(url)) or (match_obj := regexp_acg_video_bv.match(url)):
            # 匹配为投稿视频
            if "aid" in match_obj.groupdict().keys():
                avid = AId(match_obj.group("aid"))
            else:
                avid = BvId(match_obj.group("bvid"))
            title = await get_acg_video_title(session, avid)
            Logger.custom(title, Badge("投稿视频", fore="black", back="cyan"))
            acg_video_list = await get_acg_video_list(session, avid)
            # 选集过滤
            episodes = parse_episodes(args.episodes, len(acg_video_list))
            acg_video_list = list(filter(lambda item: item["id"] in episodes, acg_video_list))
            for i, acg_video_item in enumerate(acg_video_list):
                Logger.status.set("正在努力解析第 {}/{} 个视频".format(i + 1, len(acg_video_list)))
                try:
                    episode_data = await fetch_acg_video_data(
                        session, avid, i + 1, acg_video_item, args, {"title": title}, "{title}/{name}"
                    )
                except (NoAccessPermissionError, HttpStatusError, UnSupportedTypeError) as e:
                    Logger.error(e.message)
                    continue
                download_list.append(episode_data)

        elif match_obj := regexp_space_all.match(url):
            # 匹配为 UP 主个人空间
            mid = MId(match_obj.group("mid"))
            username = await get_uploader_name(session, mid)
            Logger.custom(username, Badge("UP 主投稿视频", fore="black", back="cyan"))
            acg_video_with_avid_list: list[tuple[AvId, AcgVideoListItem]] = []
            for avid in await get_uploader_space_all_videos_avids(session, mid):
                acg_video_part_list = await get_acg_video_list(session, avid)
                acg_video_with_avid_list += list(zip([avid] * len(acg_video_part_list), acg_video_part_list))
            for i, (avid, acg_video_item) in enumerate(acg_video_with_avid_list):
                Logger.status.set("正在努力解析第 {}/{} 个视频".format(i + 1, len(acg_video_with_avid_list)))
                title = await get_acg_video_title(session, avid)
                try:
                    episode_data = await fetch_acg_video_data(
                        session,
                        avid,
                        i + 1,
                        acg_video_item,
                        args,
                        {"title": title, "username": username},
                        "{username}的全部投稿视频/{title}/{name}",
                    )
                except (NoAccessPermissionError, HttpStatusError, UnSupportedTypeError) as e:
                    Logger.error(e.message)
                    continue
                download_list.append(episode_data)

        else:
            Logger.error("url 不正确～")
            sys.exit(ErrorCode.WRONG_URL_ERROR.value)

        for i, episode_data in enumerate(download_list):
            Logger.custom(
                f"{episode_data['filename']}", Badge(f"[{i+1}/{len(download_list)}]", fore="black", back="cyan")
            )
            await process_video_download(
                session,
                episode_data,
                {
                    "require_video": args.require_video,
                    "video_quality": args.video_quality,
                    "video_download_codec": args.vcodec.split(":")[0],
                    "video_save_codec": args.vcodec.split(":")[1],
                    "require_audio": args.require_audio,
                    "audio_quality": args.audio_quality,
                    "audio_download_codec": args.acodec.split(":")[0],
                    "audio_save_codec": args.acodec.split(":")[1],
                    "overwrite": args.overwrite,
                    "block_size": int(args.block_size * 1024 * 1024),
                    "num_workers": args.num_workers,
                },
            )
            Logger.print("")



import sys

from yutto.exceptions import ErrorCode

# 应当在其他模块（特别是含 3.9 语法的模块）被调用前执行
if (sys.version_info.major, sys.version_info.minor) < (3, 9):
    print("请使用 Python3.9 及以上版本哦～")
    sys.exit(ErrorCode.UNSUPPORTED_PYTHON_VERSION_ERROR.value)



from typing import NamedTuple, TypedDict

from yutto.media.codec import AudioCodec, VideoCodec
from yutto.media.quality import AudioQuality, VideoQuality
from yutto.utils.subtitle import SubtitleData
from yutto.utils.danmaku import DanmakuData


class BilibiliId(NamedTuple):
    value: str

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: "BilibiliId") -> bool:
        return self.value == other.value


class AvId(BilibiliId):
    """AID 与 BVID 的统一，大多数 API 只需要其中一种即可正常工作"""

    def to_dict(self) -> dict[str, str]:
        raise NotImplementedError("请不要直接使用 AvId")


class AId(AvId):
    """AID"""

    def to_dict(self):
        return {"aid": self.value, "bvid": ""}


class BvId(AvId):
    """BVID"""

    def to_dict(self):
        return {
            "aid": "",
            "bvid": self.value,
        }


class CId(BilibiliId):
    """视频 ID"""

    def to_dict(self):
        return {"cid": self.value}


class EpisodeId(BilibiliId):
    """番剧剧集 ID"""

    def to_dict(self):
        return {"episode_id": self.value}


class MediaId(BilibiliId):
    """番剧 ID"""

    def to_dict(self):
        return {"media_id": self.value}


class SeasonId(BilibiliId):
    """番剧（季） ID"""

    def to_dict(self):
        return {"season_id": self.value}


class MId(BilibiliId):
    """用户 ID"""

    def to_dict(self):
        return {"mid": self.value}


class VideoUrlMeta(TypedDict):
    url: str
    mirrors: list[str]
    codec: VideoCodec
    width: int
    height: int
    quality: VideoQuality


class AudioUrlMeta(TypedDict):
    url: str
    mirrors: list[str]
    codec: AudioCodec
    width: int
    height: int
    quality: AudioQuality


class MultiLangSubtitle(TypedDict):
    lang: str
    lines: SubtitleData


class EpisodeData(TypedDict):
    videos: list[VideoUrlMeta]
    audios: list[AudioUrlMeta]
    subtitles: list[MultiLangSubtitle]
    danmaku: DanmakuData
    output_dir: str
    filename: str


class DownloaderOptions(TypedDict):
    require_video: bool
    video_quality: VideoQuality
    video_download_codec: VideoCodec
    video_save_codec: str
    require_audio: bool
    audio_quality: AudioQuality
    audio_download_codec: AudioCodec
    audio_save_codec: str
    overwrite: bool
    block_size: int
    num_workers: int


if __name__ == "__main__":
    aid = AId("add")
    cid = CId("xxx")
    print("?aid={aid}&bvid={bvid}&cid={cid}".format(**aid.to_dict(), **cid.to_dict()))



import math
import random

from aiohttp import ClientSession

from yutto.typing import AvId, BvId, MId
from yutto.utils.fetcher import Fetcher


async def get_uploader_space_all_videos_avids(session: ClientSession, mid: MId) -> list[AvId]:
    ps: int = random.randint(3, 10) * 10
    space_videos_api = (
        "https://api.bilibili.com/x/space/arc/search?mid={mid}&ps={ps}&tid=0&pn={pn}&order=pubdate&jsonp=jsonp"
    )
    pn = 1
    total = 1
    all_avid: list[AvId] = []
    while pn <= total:
        space_videos_url = space_videos_api.format(mid=mid, ps=ps, pn=pn)
        json_data = await Fetcher.fetch_json(session, space_videos_url)
        total = math.ceil(json_data["data"]["page"]["count"] / ps)
        pn += 1
        all_avid += [BvId(video_info["bvid"]) for video_info in json_data["data"]["list"]["vlist"]]
    return all_avid


async def get_uploader_name(session: ClientSession, mid: MId) -> str:
    space_info_api = "https://api.bilibili.com/x/space/acc/info?mid={mid}&jsonp=jsonp"
    space_info_url = space_info_api.format(mid=mid)
    uploader_info = await Fetcher.fetch_json(session, space_info_url)
    return uploader_info["data"]["name"]



import argparse
import os
import sys
from typing import Optional

import aiohttp

from yutto.api.acg_video import (
    AcgVideoListItem,
    get_acg_video_list,
    get_acg_video_playurl,
    get_acg_video_subtitles,
    get_acg_video_title,
)
from yutto.api.bangumi import (
    BangumiListItem,
    get_bangumi_list,
    get_bangumi_playurl,
    get_bangumi_subtitles,
    get_bangumi_title,
    get_season_id_by_episode_id,
)
from yutto.api.danmaku import get_danmaku
from yutto.exceptions import ErrorCode, HttpStatusError, NoAccessPermissionError, UnSupportedTypeError
from yutto.processor.downloader import process_video_download
from yutto.processor.path_resolver import PathTemplateVariableDict, resolve_path_template
from yutto.processor.urlparser import regexp_acg_video_av, regexp_acg_video_bv, regexp_bangumi_ep
from yutto.typing import AId, AvId, BvId, EpisodeData, EpisodeId
from yutto.utils.console.logger import Badge, Logger
from yutto.utils.danmaku import EmptyDanmakuData
from yutto.utils.fetcher import Fetcher
from yutto.utils.functiontools.sync import sync


async def fetch_bangumi_data(
    session: aiohttp.ClientSession,
    episode_id: EpisodeId,
    bangumi_info: Optional[BangumiListItem],
    args: argparse.Namespace,
    subpath_variables: PathTemplateVariableDict,
    auto_subpath_template: str = "{name}",
) -> EpisodeData:
    season_id = await get_season_id_by_episode_id(session, episode_id)
    # 如果不包含详细信息，需从列表中解析
    if bangumi_info is None:
        bangumi_list = await get_bangumi_list(session, season_id)
        for bangumi_item in bangumi_list:
            if bangumi_item["episode_id"] == episode_id:
                bangumi_info = bangumi_item
                break
        else:
            Logger.error("在列表中未找到该剧集")
            sys.exit(ErrorCode.EPISODE_NOT_FOUND_ERROR.value)
    avid = bangumi_info["avid"]
    cid = bangumi_info["cid"]
    name = bangumi_info["name"]
    id = bangumi_info["id"]
    videos, audios = await get_bangumi_playurl(session, avid, episode_id, cid)
    subtitles = await get_bangumi_subtitles(session, avid, cid) if not args.no_subtitle else []
    danmaku = await get_danmaku(session, cid, args.danmaku_format) if not args.no_danmaku else EmptyDanmakuData
    # fmt: off
    subpath_variables_base: PathTemplateVariableDict = {
        "id": id,
        "name": name,
        "title": "",
        "username": "",
    }
    # fmt: on
    subpath_variables_base.update(subpath_variables)
    subpath = resolve_path_template(args.subpath_template, auto_subpath_template, subpath_variables_base)
    output_dir, filename = os.path.split(os.path.join(args.dir, subpath))
    # fmt: off
    return EpisodeData(
        videos = videos,
        audios = audios,
        subtitles = subtitles,
        danmaku = danmaku,
        output_dir = output_dir,
        filename = filename
    )
    # fmt: on


async def fetch_acg_video_data(
    session: aiohttp.ClientSession,
    avid: AvId,
    page: int,
    acg_video_info: Optional[AcgVideoListItem],
    args: argparse.Namespace,
    subpath_variables: PathTemplateVariableDict,
    auto_subpath_template: str = "{title}",
) -> EpisodeData:
    acg_video_list = await get_acg_video_list(session, avid)
    if acg_video_info is None:
        acg_video_info = acg_video_list[page - 1]
    cid = acg_video_info["cid"]
    name = acg_video_info["name"]
    id = acg_video_info["id"]
    videos, audios = await get_acg_video_playurl(session, avid, cid)
    subtitles = await get_acg_video_subtitles(session, avid, cid) if not args.no_subtitle else []
    danmaku = await get_danmaku(session, cid, args.danmaku_format) if not args.no_danmaku else EmptyDanmakuData
    # fmt: off
    subpath_variables_base: PathTemplateVariableDict = {
        "id": id,
        "name": name,
        "title": "",
        "username": "",
    }
    # fmt: on
    subpath_variables_base.update(subpath_variables)
    subpath = resolve_path_template(args.subpath_template, auto_subpath_template, subpath_variables_base)
    output_dir, filename = os.path.split(os.path.join(args.dir, subpath))
    # fmt: off
    return EpisodeData(
        videos = videos,
        audios = audios,
        subtitles = subtitles,
        danmaku = danmaku,
        output_dir = output_dir,
        filename = filename
    )
    # fmt: on


@sync
async def run(args: argparse.Namespace):
    async with aiohttp.ClientSession(
        headers=Fetcher.headers,
        cookies=Fetcher.cookies,
        trust_env=Fetcher.trust_env,
        timeout=aiohttp.ClientTimeout(total=5),
    ) as session:
        url: str = args.url
        url = await Fetcher.get_redirected_url(session, url)
        if match_obj := regexp_bangumi_ep.match(url):
            # 匹配为番剧
            episode_id = EpisodeId(match_obj.group("episode_id"))
            season_id = await get_season_id_by_episode_id(session, episode_id)
            title = await get_bangumi_title(session, season_id)
            Logger.custom(title, Badge("番剧", fore="black", back="cyan"))
            try:
                episode_data = await fetch_bangumi_data(session, episode_id, None, args, {"title": title}, "{name}")
            except (NoAccessPermissionError, HttpStatusError, UnSupportedTypeError) as e:
                Logger.error(e.message)
                sys.exit(e.code.value)

        elif (match_obj := regexp_acg_video_av.match(url)) or (match_obj := regexp_acg_video_bv.match(url)):
            # 匹配为投稿视频
            page: int = 1
            if "aid" in match_obj.groupdict().keys():
                avid = AId(match_obj.group("aid"))
            else:
                avid = BvId(match_obj.group("bvid"))
            if match_obj.group("page") is not None:
                page = int(match_obj.group("page"))
            title = await get_acg_video_title(session, avid)
            Logger.custom(title, Badge("投稿视频", fore="black", back="cyan"))
            try:
                episode_data = await fetch_acg_video_data(
                    session,
                    avid,
                    page,
                    None,
                    args,
                    {"title": title},
                    "{title}",
                )
            except (NoAccessPermissionError, HttpStatusError, UnSupportedTypeError) as e:
                Logger.error(e.message)
                sys.exit(e.code.value)

        else:
            Logger.error("url 不正确～")
            sys.exit(ErrorCode.WRONG_URL_ERROR.value)

        await process_video_download(
            session,
            episode_data,
            {
                "require_video": args.require_video,
                "video_quality": args.video_quality,
                "video_download_codec": args.vcodec.split(":")[0],
                "video_save_codec": args.vcodec.split(":")[1],
                "require_audio": args.require_audio,
                "audio_quality": args.audio_quality,
                "audio_download_codec": args.acodec.split(":")[0],
                "audio_save_codec": args.acodec.split(":")[1],
                "overwrite": args.overwrite,
                "block_size": int(args.block_size * 1024 * 1024),
                "num_workers": args.num_workers,
            },
        )



from typing import Any, Literal, Union
from enum import Enum


class ErrorCode(Enum):

    UNSUPPORTED_PYTHON_VERSION_ERROR = 9
    HTTP_STATUS_ERROR = 10
    NO_ACCESS_PERMISSION_ERROR = 11
    UNSUPPORTED_TYPE_ERROR = 12
    WRONG_ARGUMENT_ERROR = 13
    WRONG_URL_ERROR = 14
    EPISODE_NOT_FOUND_ERROR = 15


class SuccessCode(Enum):
    SUCCESS = 0


ReturnCode = Union[ErrorCode, SuccessCode]


class YuttoBaseException(Exception):
    code: ErrorCode
    message: str

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


class HttpStatusError(YuttoBaseException):
    code = ErrorCode.HTTP_STATUS_ERROR


class NoAccessPermissionError(YuttoBaseException):
    code = ErrorCode.NO_ACCESS_PERMISSION_ERROR


class UnSupportedTypeError(YuttoBaseException):
    code = ErrorCode.UNSUPPORTED_TYPE_ERROR


if __name__ == "__main__":
    import sys

    try:
        raise HttpStatusError("HTTP 错误")
    except (HttpStatusError, UnSupportedTypeError) as e:
        print(e.code.value, e.message)
        sys.exit(e.code.value)



import aiohttp

from yutto.typing import CId
from yutto.utils.fetcher import Fetcher
from yutto.utils.danmaku import DanmakuData, DanmakuSaveType


async def get_xml_danmaku(session: aiohttp.ClientSession, cid: CId) -> str:
    danmaku_api = "http://comment.bilibili.com/{cid}.xml"
    return await Fetcher.fetch_text(session, danmaku_api.format(cid=cid), encoding="utf-8")


async def get_protobuf_danmaku(session: aiohttp.ClientSession, cid: CId, segment_id: int = 1) -> bytes:
    danmaku_api = "http://api.bilibili.com/x/v2/dm/web/seg.so?type=1&oid={cid}&segment_index={segment_id}"
    return await Fetcher.fetch_bin(session, danmaku_api.format(cid=cid, segment_id=segment_id))


async def get_danmaku(
    session: aiohttp.ClientSession,
    cid: CId,
    save_type: DanmakuSaveType,
    last_n_segments: int = 2,
) -> DanmakuData:
    # 暂时默认使用 XML 源
    source_type = "xml" if save_type == "xml" or save_type == "ass" else "protobuf"
    # fmt: off
    danmaku_data: DanmakuData = {
        "source_type": source_type,
        "save_type": save_type,
        "data": []
    }
    # fmt: on

    if source_type == "xml":
        danmaku_data["data"].append(await get_xml_danmaku(session, cid))
    else:
        for i in range(1, last_n_segments + 1):
            danmaku_data["data"].append(await get_protobuf_danmaku(session, cid, i))
    return danmaku_data



# type: ignore
class Singleton(type):
    """单例模式元类

    @refs: https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python

    # Usage
    ```
    class MyClass(BaseClass, metaclass=Singleton):
        pass
    ```
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yutto',
 'yutto.api',
 'yutto.cli',
 'yutto.media',
 'yutto.processor',
 'yutto.utils',
 'yutto.utils.console',
 'yutto.utils.functiontools']

package_data = \
{'': ['*']}

install_requires = \
['aiofiles==0.7.0', 'aiohttp==3.7.4.post0', 'biliass==1.3.4']

entry_points = \
{'console_scripts': ['yutto = yutto.__main__:main']}

setup_kwargs = {
    'name': 'yutto',
    'version': '2.0.0b1',
    'description': 'yutto 一个可爱且任性的 B 站视频下载器',
    'long_description': '# yutto<sup>2.0.0-beta</sup>\n\n<p align="center">\n   <a href="https://python.org/" target="_blank"><img alt="PyPI - Python Version" src="https://img.shields.io/pypi/pyversions/yutto?logo=python&style=flat-square"></a>\n   <a href="https://pypi.org/project/yutto/" target="_blank"><img src="https://img.shields.io/pypi/v/yutto?style=flat-square" alt="pypi"></a>\n   <a href="https://pypi.org/project/yutto/" target="_blank"><img alt="PyPI - Downloads" src="https://img.shields.io/pypi/dm/yutto?style=flat-square"></a>\n   <a href="LICENSE"><img alt="LICENSE" src="https://img.shields.io/github/license/SigureMo/yutto?style=flat-square"></a>\n   <a href="https://gitmoji.dev"><img src="https://img.shields.io/badge/gitmoji-%20😜%20😍-FFDD67?style=flat-square" alt="Gitmoji"></a>\n</p>\n\nyutto，一个可爱且任性的 B 站下载器（CLI）\n\n## 版本号为什么是 2.0\n\n因为 yutto 是 bilili 的後輩呀～\n\n## 安装预览版\n\n在此之前请确保安装 Python3.9（不支持 3.8 及以下，3.10 尚处于 beta，没有测试）与 FFmpeg（参照 [bilili 文档](https://bilili.sigure.xyz/guide/getting-started.html)）\n\n当前 yutto 尚处于 beta 阶段，有任何建议尽管提出～～～\n\n### pip 安装\n\n```bash\npip install --pre yutto\n```\n\n### git clone\n\n```bash\ngit clone https://github.com/SigureMo/yutto.git\npip install poetry\npoetry build\npip install ./dist/yutto-*.whl\n```\n\n## 主要功能\n\n### 基本命令\n\nyutto 的基本命令如下：\n\n```bash\nyutto <url>\n```\n\n你可以通过 `yutto -h` 查看详细命令参数。\n\n如果你需要下载**单个**视频，只需要使用 yutto 加上这个视频的地址即可。它支持 av/BV 号以及相应带 p=n 参数的投稿视频页面，也支持 ep 号（episode_id）的番剧页面。\n\n比如只需要这样你就可以下载[《転スラ日記》](https://www.bilibili.com/bangumi/play/ep395211)第一话：\n\n```bash\nyutto https://www.bilibili.com/bangumi/play/ep395211\n```\n\nyutto 还支持直接使用能够唯一定位资源的 id 来作为 `<url>`，刚刚的功能与下面的简化后的命令功能是完全一样的\n\n```bash\nyutto ep395211\n```\n\n不过有时你可能想要批量下载很多剧集，因此 yutto 提供了用于批量下载的参数 `-b/--batch`，它不仅支持前面所说的单个视频所在页面地址（会解析该单个视频所在的系列视频），还支持一些明确用于表示系列视频的地址，比如 md 页面（media_id）、ss 页面（season_id）。\n\n比如像下面这样就可以下载《転スラ日記》所有已更新的剧集：\n\n```bash\nyutto --batch https://www.bilibili.com/bangumi/play/ep395211\n```\n\n### 基础参数\n\n> 大部分参数与 bilili 重合，可参考 [bilili 的 cli 文档](https://bilili.nyakku.moe/cli/)\n\nyutto 支持一些基础参数，无论是批量下载还是单视频下载都适用。\n\n<details>\n<summary>点击展开详细参数</summary>\n\n#### 最大并行 worker 数量\n\n-  参数 `-n` 或 `--num-workers`\n-  默认值 `8`\n\n与 bilili 不同的是，yutto 并不是使用多线程实现并行下载，而是使用协程实现的，本参数限制的是最大的并行 Worker 数量。\n\n#### 指定视频清晰度等级\n\n-  参数 `-q` 或 `--video-quality`\n-  可选值 `125 | 120 | 116 | 112 | 80 | 74 | 64 | 32 | 16`\n-  默认值 `125`\n\n清晰度对应关系如下\n\n<!-- prettier-ignore -->\n|code|清晰度|\n|:-:|:-:|\n|125|HDR 真彩|\n|120|4K 超清|\n|116|1080P 60帧|\n|112|1080P 高码率|\n|80|1080P 高清|\n|74|720P 60帧|\n|64|720P 高清|\n|32|480P 清晰|\n|16|360P 流畅|\n\n并不是说指定某个清晰度就一定会下载该清晰度的视频，yutto 只会尽可能满足你的要求，如果不存在指定的清晰度，yutto 就会按照默认的清晰度搜索机制进行调节，比如指定清晰度为 `80`，**首先会依次降清晰度搜索** `74`、`64`、`32`、`16`，如果依然找不到合适的则**继续升清晰度搜索** `112`、`116`、`120`、`125`。\n\n#### 指定音频码率等级\n\n-  参数 `-aq` 或 `--audio-quality`\n-  可选值 `30280 | 30232 | 30216`\n-  默认值 `30280`\n\n码率对应关系如下\n\n<!-- prettier-ignore -->\n|code|码率|\n|:-:|:-:|\n|30280|320kbps|\n|30232|128kbps|\n|30216|64kbps|\n\n清晰度自动调节机制与视频清晰度一致，也采用先降后升的匹配机制。\n\n#### 指定视频编码\n\n-  参数 `--vcodec`\n-  下载编码可选值 `"hevc" | "avc"`\n-  保存编码可选值 FFmpeg 所有可用的视频编码器\n-  默认值 `"avc:copy"`\n\n该参数略微复杂，前半部分表示在下载时**优先**选择哪一种编码的视频流，后半部分则表示在合并时如何编码视频流，两者使用 `:` 分隔。\n\n值得注意的是，前半的下载编码只是优先下载的编码而已，如果不存在该编码，则仍会像视频清晰度调节机制一样自动选择其余编码。\n\n而后半部分的参数如果设置成非 `copy` 的值则可以确保在下载完成后对其进行重新编码，而且不止支持 `hevc` 与 `avc`，只要你的 FFmpeg 支持的视频编码器，它都可以完成。\n\n#### 指定音频编码\n\n-  参数 `--acodec`\n-  下载编码可选值 `"mp4a"`\n-  保存编码可选值 FFmpeg 所有可用的音频编码器\n-  默认值 `"mp4a:copy"`\n\n详情同视频编码。\n\n#### 仅下载视频流\n\n-  参数 `--video-only`\n-  默认值 `False`\n\n#### 仅下载音频流\n\n-  参数 `--audio-only`\n-  默认值 `False`\n\n仅下载其中的音频流，保存为 `.aac` 文件。\n\n值得注意的是，在不选择视频流时，嵌入字幕、弹幕功能将无法工作。\n\n#### 弹幕格式选择\n\n-  参数 `-df` 或 `--danmaku-format`\n-  可选值 `"ass" | "xml" | "protobuf"`\n-  默认值 `"ass"`\n\nB 站提供了 `xml` 与 `protobuf` 两种弹幕数据接口，yutto 会自动下载 `xml` 格式弹幕并转换为 `ass` 格式，如果你不喜欢 yutto 自动转换的效果，可以选择输出格式为 `xml` 或 `protobuf`，手动通过一些工具进行转换，比如 yutto 和 bilili 所使用的 [biliass](https://github.com/ShigureLab/biliass)，或者使用 [us-danmaku](https://tiansh.github.io/us-danmaku/bilibili/) 进行在线转换。\n\n如果你不想下载弹幕，只需要使用参数 `--no-danmaku` 即可。\n\n#### 下载块大小\n\n-  参数 `-bs` 或 `--block-size`\n-  默认值 `0.5`\n\n以 MiB 为单位，为分块下载时各块大小，不建议更改。\n\n#### 强制覆盖已下载文件\n\n-  参数 `-w` 或 `--overwrite`\n-  默认值 `False`\n\n#### 代理设置\n\n-  参数 `-x` 或 `--proxy`\n-  可选值 `"auto" | "no" | <https?://url/to/proxy/server>`\n-  默认值 `"auto"`\n\n设置代理服务器，默认是从环境变量读取，`no` 则为不设置代理，设置其它 http/https url 则将其作为代理服务器。\n\n#### 存放根目录\n\n-  参数 `-d` 或 `--dir`\n-  默认值 `"./"`\n\n#### 存放子路径模板\n\n-  参数 `-tp` 或 `--subpath-template`\n-  可选参数变量 `title | id | name | username` （以后可能会有更多）\n-  默认值 `"{auto}"`\n\n通过配置子路径模板可以灵活地控制视频存放位置。\n\n默认情况是由 yutto 自动控制存放位置的。比如下载单个视频时默认就是直接存放在设定的根目录，不会创建一层容器目录，此时自动选择了 `{name}` 作为模板；而批量下载时则会根据视频层级生成多级目录，比如番剧会是 `{title}/{name}`，首先会在设定根目录里生成一个番剧名的目录，其内才会存放各个番剧剧集视频，这样方便了多个不同番剧的管理。当然，如果你仍希望将番剧直接存放在设定根目录下的话，可以修改该参数值为 `{name}`即可。\n\n另外，该功能语法由 Python format 函数模板语法提供，所以也支持一些高级的用法，比如 `{id:0>3}{name}`。\n\n值得注意的是，并不是所有变量在各种场合下都会提供，比如 `username` 变量当前仅在 UP 主全部投稿视频才提供，在其它情况下不应使用它。\n\n#### url 别名文件路径\n\n-  参数 `-af` 或 `--alias-file`\n-  默认值 `None`\n\n指定别名文件路径，别名文件中存放一个别名与其对应的 url，使用空格或者 `=` 分隔，示例如下：\n\n```\nrimuru1=https://www.bilibili.com/bangumi/play/ss25739/\nrimuru2=https://www.bilibili.com/bangumi/play/ss36170/\nrimuru-nikki=https://www.bilibili.com/bangumi/play/ss38221/\n```\n\n比如将上述文件存储到 `~/.yutto_alias`，则通过以下命令即可解析该文件：\n\n```bash\nyutto rimuru1 --batch --alias-file=\'~/.yutto_alias\'\n```\n\n当参数值为 `-` 时，会从标准输入中读取：\n\n```bash\ncat ~/.yutto_alias | yutto rimuru-nikki --batch --alias-file -\n```\n\n#### Cookies 设置\n\n-  参数 `-c` 或 `--sessdata`\n-  默认值 `""`\n\n设置 Cookies 后你才可以下载更高清晰度以及更多的剧集，当你传入你的大会员 `SESSDATA` 时（当然前提是你是大会员），你就可以下载大会员可访问的资源咯。\n\n<details><summary> SESSDATA 获取方式 </summary>\n\n这里用 Chrome 作为示例，其它浏览器请尝试类似方法。\n\n首先，用你的帐号登录 B 站，然后随便打开一个 B 站网页，比如[首页](https://www.bilibili.com/)。\n\n按 F12 打开开发者工具，切换到 Network 栏，刷新页面，此时第一个加载的资源应该就是当前页面的 html，选中该资源，在右侧 「Request Headers」 中找到 「cookie」，在其中找到类似于 `SESSDATA=d8bc7493%2C2843925707%2C08c3e*81;` 的一串字符串，复制这里的 `d8bc7493%2C2843925707%2C08c3e*81`，这就是你需要的 `SESSDATA`。\n\n</details>\n\n另外，由于 SESSDATA 中可能有特殊符号，所以传入时你可能需要使用双引号来包裹\n\n```bash\nyutto <url> -c "d8bc7493%2C2843925707%2C08c3e*81"\n```\n\n当然，示例里的 SESSDATA 是无效的，请使用自己的 SESSDATA。\n\n#### 不下载弹幕\n\n-  参数 `--no-danmaku`\n-  默认值 `False`\n\n#### 不下载字幕\n\n-  参数 `--no-subtitle`\n-  默认值 `False`\n\n#### 不显示颜色\n\n-  参数 `--no-color`\n-  默认值 `False`\n\n#### 启用 Debug 模式\n\n-  参数 `--debug`\n-  默认值 `False`\n\n</details>\n\n### 批量参数\n\n有些参数是只有批量下载时才可以使用的\n\n<details>\n<summary>点击展开详细参数</summary>\n\n#### 启用批量下载\n\n-  参数 `-b` 或 `--batch`\n-  默认值 `False`\n\n只需要 `yutto --batch <url>` 即可启用批量下载功能。\n\n#### 选集\n\n-  参数 `-p` 或 `--episodes`\n-  默认值 `^~$`\n\n-  参数 `-p` 或 `--episodes`\n-  默认值 `^~$`（也即全选）\n\n也就是选集咯，其语法是这样的\n\n-  `<p1>` 单独下某一剧集\n   -  支持负数来选择倒数第几话\n   -  此外还可以使用 `^` 与 `$` 来分别代表 `1` 与 `-1`\n-  `<p_start>~<p_end>` 使用 `~` 可以连续选取\n-  `<p1>,<p2>,<p3>,...,<pn>` 使用 `,` 可以不连续选取\n\nemmm，直接看的话大概并不能知道我在说什么，所以我们通过几个小例子来了解其语法\n\n```bash\n# 假设要下载一个具有 24 话的番剧\n# 如果我们只想下载第 3 话，只需要这样\nyutto <url> -b -p 3\n# 那如果我想下载第 5 话到第 7 话呢，使用 `~` 可以连续选中\nyutto <url> -b -p 5~7\n# 那我想下载第 12 话和第 17 话又要怎么办？此时只需要 `,` 就可以将多个不连续的选集一起选中\nyutto <url> -b -p 12,17\n# 那我突然又想将刚才那些都选中了呢？还是使用 `,` 呀，将它们连在一起即可\nyutto <url> -b -p 3,5~7,12,17\n# 嗯，你已经把基本用法都了解过了，很简单吧～\n# 下面是一些语法糖，不了解也完全不会影响任何功能哒～\n# 那如果我只知道我想下载倒数第 3 话，而不想算倒数第三话是第几话应该怎么办？\n# 此时可以用负数哒～不过要注意的是，开头如果是 `-` 的话前面应该使用 `=`\nyutto <url> -b -p=-3\n# 那么如果想下载最后一话你可能会想到 `-p=-1` 对吧？不过我内置了两个符号分别代表第一话（^）和最后一话（$）\n# 像下面这样就可以直接下载最后一话啦～\nyutto <url> -b -p $\n# 所有语法都了解完啦，我们看一个稍微复杂的例子\nyutto <url> -b -p ^~3,10,12~14,16,-4~$\n# 很明显，上面的例子就是下载前 3 话、第 10 话、第 12 到 14 话、第 16 话以及后 4 话\n```\n\n下面是一些要注意的问题\n\n1. 这里使用的序号是视频的顺序序号，而不是番剧所标注的`第 n 话`，因为有可能会出现 `第 x.5 话` 等等的特殊情况，此时一定要按照顺序自行计数。\n2. 参数值里一定不要加空格\n3. 参数值开头为 `-` 时前面应该使用 `=` 而非空格\n\n#### 同时下载附加剧集\n\n-  参数 `-s` 或 `--with-section`\n-  默认值 `False`\n\n</details>\n\n## 从 bilili1.x 迁移\n\n### 取消的功能\n\n-  `- bilibili` 目录的生成\n-  播放列表生成\n-  源格式修改功能（不再支持 flv 源视频下载，如果仍有视频不支持 dash 源，请继续使用 bilili）\n-  对 Python3.8 的支持，最低支持 Python3.9\n-  取消下载询问\n\n### 默认行为的修改\n\n-  使用协程而非多线程进行下载，同时也不是批量解析批量下载，而是边解析边下载\n-  默认生成弹幕为 ASS\n-  默认启用从多镜像源下载的特性\n-  不仅可以控制是否使用系统代理，还能配置特定的代理服务器\n\n### 新增的特性\n\n-  单视频下载与批量下载命令分离（`bilili` 命令与 `yutto --batch` 相类似）\n-  音频/视频编码选择\n-  仅下载音频/视频\n-  存放子路径的自由定制\n-  支持 url alias\n-  支持 file scheme\n-  更多的批下载支持（现已支持 UP 主全部视频下载，扩展其它批下载支持也很简单）\n-  更加完善的 warning 与 error 提示\n-  支持仅输入 id 即可下载（aid、bvid、episode_id 等）\n\n## 小技巧\n\n### 使用 uvloop 提升协程效率\n\n貌似 uvloop 可以提高协程效率，如果你是非 Windows 的话，可以试一下安装 uvloop：\n\n```bash\npip install uvloop\n```\n\n具体提升多少我也不太清楚啦，没有测过\n\n### 使用 url alias\n\nyutto 新增的 url alias 可以让你下载正在追的番剧时不必每次都打开浏览器复制 url，只需要将追番列表存储在一个文件中，并为这些 url 起一个别名即可\n\n```\nrimuru-nikki=https://www.bilibili.com/bangumi/play/ss38221/\n```\n\n之后下载最新话只需要\n\n```\nyutto --batch rimuru-nikki --alias-file=/path/to/alias-file\n```\n\n### 使用文件列表\n\n现在 url 不仅支持 http/https 链接与裸 id，还支持使用文件路径与 file scheme 来用于表示文件列表，文件列表以行分隔，每行写一次命令的参数，该参数会覆盖掉主程序中所使用的参数，示例如下：\n\n首先将下面的文件存储到一个地方\n\n```\nhttps://www.bilibili.com/bangumi/play/ss38221/ --batch -p $\nhttps://www.bilibili.com/bangumi/play/ss38260/ --batch -p $\n```\n\n然后运行\n\n```\nyutto file:///path/to/list\n```\n\n即可分别下载这两个番剧的最新一话\n\n或者直接使用相对或者绝对路径也是可以的\n\n```\nyutto ./path/to/list\n```\n\n值得注意的是，在文件列表各项里的参数优先级是高于命令里的优先级的，比如文件中使用：\n\n```\nrimuru1 --batch -p $ --no-danmaku --vcodec="hevc:copy"\nrimuru2 --batch -p $\n```\n\n而命令中则使用\n\n```\nyutto file:///path/to/list --vcodec="avc:copy"\n```\n\n最终下载的 rimuru1 会是 "hevc:copy"，而 rimuru2 则会是 "avc:copy"\n\n另外，文件列表也是支持 alias 的，你完全可以为该列表起一个别名，而且列表支持是支持嵌套的（虽然没什么用 2333）\n\n## FAQ\n\n### 名字的由来\n\n[《転スラ日記》第一话](https://www.bilibili.com/bangumi/play/ep395211) 00:25\n\n### yutto 会很快替代 bilili 吗\n\n短期内不会，bilili 并不会消失，在一段时间内 bilili 仍会做纠错性维护，只是不会提供新特性了。\n\nyutto 现在也还不是非常稳定，需要稳定的体验的话请继续使用 bilili ～\n\n## Roadmap\n\n### 2.0.0-beta\n\n-  [x] feat: 支持 bare name (bare id, bare path)\n-  [x] refactor: url 列表能够预线性展开\n-  [x] feat: 添加各种 return code\n-  [x] test: 编写单元测试\n\n### future\n\n-  [ ] refactor: 尽可能使下载器成为独立模块（也许？）\n-  [ ] feat: 字幕、弹幕嵌入视频支持（也许？）\n-  [ ] refactor: 以插件形式支持更多音视频处理方面的功能，比如 autosub（也许？）\n-  [ ] feat: 更多批下载支持（UP 主、收藏夹、稍后再看等）\n-  [ ] docs: 静态文档\n-  [ ] feat: 以及更加可爱～\n\n## 参考\n\n-  基本结构：<https://github.com/SigureMo/bilili>\n-  协程下载：<https://github.com/changmenseng/AsyncBilibiliDownloader>\n-  弹幕转换：<https://github.com/ShigureLab/biliass>\n-  样式设计：<https://github.com/willmcgugan/rich>\n\n## 参与贡献\n\n请阅读 [CONTRIBUTING.md](CONTRIBUTING.md)\n',
    'author': 'Nyakku Shigure',
    'author_email': 'sigure.qaq@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/SigureMo/yutto',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9.0,<4.0.0',
}


setup(**setup_kwargs)
